/**   @file  oomstub.c
 *   @brief  C memory management stubs
 *
 *  This file implements fake malloc, free and realloc functions which can
 *  fail and simulate OOM after a countdown.
 *  This library is not threadsafe.
 *  Inspired from https://github.com/bkthomps/Containers test suite
 *
 *  @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *
 *  @internal
 *       Created:  10/12/2024
 *      Revision:  none
 * Last modified:  2024-12-11 19:45
 *      Compiler:  gcc
 *  Organization:  Cerbelle.net
 *     Copyright:  Copyright (c) 2024, François Cerbelle
 *
 *  This source code is released for free distribution under the terms of the
 *  GNU General Public License as published by the Free Software Foundation.
 */

#include "liboom/oomstub.h"
#include <stdlib.h>                             /* NULL */
#include <dlfcn.h>

#ifndef RTLD_NEXT
#define RTLD_NEXT ((void *) -1L)
#endif

/** Countdown before malloc failure
 *  The value is decremented at each malloc call. I use a single countdown
 *  shared between all stub functions because the compiler can optimize the code
 *  and choose to directly use a malloc call instead of a realloc(NULL,xyz).
 */
static int oomstub_countdown = 0;

/** Functor to store the original malloc function
 *  This functor saves the original malloc function to call if no failure is
 *  expected.
 */
static void *(*real_malloc)(size_t);

/** Functor to store the original calloc function
 *  This functor saves the original calloc function to call if no failure is
 *  expected.
 */
static void *(*real_calloc)(size_t, size_t);

/** Functor to store the original realloc function
 *  This functor saves the original realloc function to call if no failure is
 *  expected.
 */
static void *(*real_realloc)(void *, size_t);

/** Sets the malloc countdown before triggering a failure
 *  @param  counter Value to set in the internal countdown
 *  @see malloc
 *  @see calloc
 *  @see realloc
 */
void oomstub_setcountdown(const int counter) {
    oomstub_countdown = counter;
}

/** Gets the current malloc countdown before triggering a failure
 *  @return The current internal countdown value
 *  @see malloc
 *  @see calloc
 *  @see realloc
 */
int  oomstub_getcountdown() {
    return oomstub_countdown;
}

/** stdc malloc stub function
 *  @param  size Size in bytes to try to allocate
 *  @return Pointer to the allocated heap block
 *      @retval NULL The allocation failed
 *
 *  This function is called instead of the upstream malloc function. If not
 *  already done, it creates a backup ot the upstream function pointer. If the
 *  countdown equals 1, it simulates OOM and returns NULL. Otherwise, it calls
 *  the upstream malloc and returns its return value. Each call to this stub
 *  function decrements the countdown until it reaches 0.
 */
void *malloc(size_t size) {
    if (!real_malloc) {
        *(void **) (&real_malloc) = dlsym(RTLD_NEXT, "malloc");
    }
    if (oomstub_countdown == 1) {
        oomstub_countdown = 0;
        return NULL;
    }
    if (oomstub_countdown > 0) {
        oomstub_countdown--;
    }
    return real_malloc(size);
}

/** stdc calloc stub function
 *  @param  count How many contiguous records to allocate
 *  @param  size Size in bytes of a single record
 *  @return Pointer to the allocated heap block
 *      @retval NULL The allocation failed
 *
 *  This function is called instead of the upstream calloc function. If not
 *  already done, it creates a backup ot the upstream function pointer. If the
 *  countdown equals 1, it simulates OOM and returns NULL. Otherwise, it calls
 *  the upstream calloc and returns its return value. Each call to this stub
 *  function decrements the countdown until it reaches 0.
 */
void *calloc(size_t count, size_t size) {
    if (!real_calloc) {
        *(void **) (&real_calloc) = dlsym(RTLD_NEXT, "calloc");
    }
    if (oomstub_countdown == 1) {
        oomstub_countdown = 0;
        return NULL;
    }
    if (oomstub_countdown > 0) {
        oomstub_countdown--;
    }
    return real_calloc(count, size);
}

/** stdc realloc stub function
 *  @param  ptr Pointer to the dynamic heap block to resize
 *  @param  new_size New size in bytes to request
 *  @return Pointer to the reallocated heap block
 *      @retval NULL The allocation failed (important details in description)
 *
 *  This function is called instead of the upstream realloc function. If not
 *  already done, it creates a backup ot the upstream function pointer. If the
 *  countdown equals 1, it simulates OOM and returns NULL. Otherwise, it calls
 *  the upstream realloc and returns its return value. Each call to this stub
 *  function decrements the countdown until it reaches 0.
 *
 *  @note The failure countdown should not apply when shrinking the size. It is
 *  meaningless as it can not fail in real life. It would need bookkeeping to
 *  detect arbitrary shrinks. Thus, this stub function also counts shrinking
 *  calls in the countdown and apply failures.
 *  The tradeof is that the function returns NULL in case of a simulated failure
 *  and a pointer in case of success for any case but not for the "free" case.
 *  In the "free" case (new_size==0), it returns a not-null value in case of
 *  failure and NULL in case of success.
 *
 *  @warning if the compiler optimizes a realloc(ptr,0) with a direct free(ptr)
 *  call, the countdown is not decremented. I can not detect this condition and
 *  I have no tradeoff/workaround. Diverting free() would decrement the
 *  countdown when not expected.
 */
void *realloc(void *ptr, size_t new_size) {
    if (!real_realloc) {
        *(void **) (&real_realloc) = dlsym(RTLD_NEXT, "realloc");
    }
    if (oomstub_countdown == 1) {
        oomstub_countdown = 0;
        if (new_size > 0)
            return NULL;
        else
            return ptr;
    }
    if (oomstub_countdown > 0) {
        oomstub_countdown--;
    }

    return real_realloc(ptr, new_size);
}

/* vim: set tw=80: */
